%% Paradigme de Programare
%% Laboratorul 12
%% Probleme de satisfacere a restricțiilor

%% Problemele de satisfacere a restricțiilor sunt definite prin
%% tupluri <Variables, Domains, Constraints>.

%% Vom reprezenta o restricție printr-o structură c(Vars, Relation)
%% unde Vars va reprezenta lista variabilelor implicate în restricție, iar
%% Relation va reprezenta un scop ce implică acele variabile. De exemplu,
%% o restricție pentru a specifica că două variabile sunt diferite:
%% c([X1, X2], X1 =\= X2).

%% Dorim să implementăm un mecanism general de rezolvare a problemlor de
%% satisfacere a restricțiilor, pornind de la un backracking simplu și,
%% adăugate pe rând, verificare înainte, precum și ordonare a variabilelor
%% pentru a micșora spațiul de căutare.

%% 1. Reprezentarea problemelor

%% Problema colorării hărților

map_coloring([V1, V2, V3, V4, V5],
             [[r], [r,g,b], [r,g,b], [r,g,b], [r,g,b]],
             [c([V1,V2], V1 \== V2), c([V1,V3], V1 \== V3),
              c([V2,V3], V2 \== V3), c([V1,V4], V1 \== V4),
              c([V1,V5], V1 \== V5), c([V4,V5], V1 \== V5),
              c([V3,V4], V3 \== V4)]).

%% 1.a. [1 punct]
%% Scrieți un predicat four_queens(-Vars, -Domains, -Constrains)
%% prin care să reprezentați problema celor patru regine.

four_queens(_Vars, _Domains, _Constraints):-fail.

%% 1.b. [2 puncte]
%% Scrieți un predicat n_queens(+N, -Vars, -Domains, -Constrains)
%% prin care să construiți reprezentarea pentru problema celor N
%% regine, unde N este orice număr natural.

n_queens(_N, _Vars, _Domains, _Constraints):-fail.

%% 1.c. [1 punct]
%% Scrieți un predicat cryptharithm(-Vars, -Domains, -Constraints)
%% prin care să reprezentați următorul puzzle criptaritmetic:
%%     NO+
%%    GUN
%%     NO
%% ------
%%   HUNT
%%
%% Dacă scrieți o singură constrângere, nu veți putea putea micșora spațiul
%% de căutare pe parcurs. Încercați să scrieți mai multe constrângeri ce
%% implică un număr mai mic de variabile.

cryptharithm(_Vars, _Domains, _Constraints):-fail.


%% 2. Rezolvarea folosind backtracking
%% Observați mecanismul de rezolvare a unei astfel de probleme:
%% bkt(+Vars, +Domains, +Constraints)
%% Un scop bkt va reuși prin legarea variabilelor din Vars la valori din
%% Domains ce respectă restricțiile din lista Constraints.

bkt([], [], Constraints):-
        check_constraints(Constraints).
bkt([Var|OtherVars], [Dom|OtherDomains], Constraints):-
        member(Var, Dom),
        bkt(OtherVars, OtherDomains, Constraints).

%% 2.1 [1 punct]
%% Scrieți predicatul check_constraints(+Constraints)
%% care verifică dacă toate constrângerile din lista ce reprezintă
%% singurul argument sunt adevărate. Se presupune că toate variabilele
%% implicate în constrângeri sunt legate.

check_constraints(_Constraints):-fail.

%% Testare 1 & 2:

solve_bkt(Problem):-
        call(Problem, Vars, Domains, Constraints),
        bkt(Vars, Domains, Constraints).

check_12:-
        findall(a, solve_bkt(map_coloring), A), length(A, 4), write('.'),
        findall(a, solve_bkt(four_queens), B), length(B, 2), write('.'),
        findall(a, solve_bkt(cryptharithm), C), length(C, 1), write('.'),
        findall(a, solve_bkt(n_queens(1)), N1), length(N1, 1), write('.'),
        findall(a, solve_bkt(n_queens(2)), N2), length(N2, 0), write('.'),
        findall(a, solve_bkt(n_queens(3)), N3), length(N3, 0), write('.'),
        findall(a, solve_bkt(n_queens(4)), N4), length(N4, 2), write('.'),
        findall(a, solve_bkt(n_queens(5)), N5), length(N5, 10), write('.'),
        findall(a, solve_bkt(n_queens(6)), N6), length(N6, 4), write('.'),
        findall(a, solve_bkt(n_queens(7)), N7), length(N7, 40), write('.'),
        !, write('Exercițiile 1 și 2 rezolvate corect'), nl.

%% 3. Revizuirea domeniului unei variabile
%% Pentru acest exercițiu presupunem că avem doar restricții binare (care
%% implică doar perechi de variabile, precum în problema celor n regine sau
%% problema colorării hărților).
%%
%% 3. [1 punct]
%% Scrieți un predicat revise(+Var, +OldDomain, +Constraint, -NewDomain)
%% la satisfacerea căruia lista NewDomain să conțină acele valori din
%% OldDomain pentru care variabila Var respectă restricția Constraint.
%% Vom presupune că cea de-a doua variabilă a fost fixată.
%% Nu folosiți findall/3 sau bagof/3!

revise(_Var, _OldDomain, _Constraint, _NewDomain):-fail.

%% Testare 3

check_3:-
        revise(X,[1,2,3,4,5],c([X,3], X + 3 > 4),RDom1),
        sort(RDom1,RDom1s), RDom1s == [2,3,4,5], var(X), write('.'),
        revise(X,[1,2,3,4,5,6,7],c([5,X], (X + 2) mod 3 =:= 1), RDom2),
        sort(RDom2,RDom2s), RDom2s == [2,5], var(X), write('.'),
        revise(X,[1,2,3,4,5,6,7,8,9],c([5,X], X < 0), RDom3),
        RDom3 == [], var(X), write('.'),
        write('Exercițiul 3 rezolvat corect'), nl.

%% 4. Căutare cu verificare înainte

%% 4.1. [1 punct]
%% Scrieți un predicat
%% get_constraints(+Var, +AllConstraints, -VarConstraints, -Other)
%% pe care să-l folosiți pentru a lega variabila VarConstraints la acele
%% constrângeri care privesc variabila Var și o altă variabilă neinițializată.
%% Lista Other va conține toate celelalte constrângeri din AllConstraints.
%% Folosiți predicatul var/1 pentru a testa dacă o variabilă este nelegată.

get_constraints(_Var, _AllConstraints, _VarConstraints, _Other):-fail.

%% Testare 4.1

check_41:-
        get_constraints(X1,
                        [c([X1,X2], X1 > X2), c([X1,3], X1 > 3),
                         c([X3,X1], X1 < X3), c([X3,X2], X3 > X2),
                         c([2,X1], X1 < 2)],
                        CVar, COther),
        var(X1), var(X2), var(X3),
        sort(CVar, CVarS), CVarS == [c([X1,X2],X1>X2),c([X3,X1],X1<X3)],
        sort(COther, COtherS),
        COtherS == [c([X1,3], X1>3),c([X3,X2],X3>X2),c([2,X1],X1<2)],
        write('.'),!,write('Exercițiul 4.1 rezolvat corect!'), nl.

%% 4.2. [3 puncte]
%% Scrieți predicatul forward_checking(+Vars, +Domains, +Constraints) plecând
%% de la predicatul bkt/3 scris mai sus, astfel încât după fixarea valorii
%% unei variabile să se revizuiască domeniile de valori ale celorlalte
%% variabile ce sunt implicate în restricții binare cu aceasta.
%% Dacă unul dintre aceste domenii conține zero elemente, nu se va
%% trece la următoarea variabilă.

forward_checking(_Vars, _Domains, _Constraints):-fail.

%% Testare 4.2

solve_fc(Problem):-
        call(Problem, Vars, Domains, Constraints),
        forward_checking(Vars, Domains, Constraints).

check_42:-
        findall(a, solve_fc(map_coloring), A), length(A, 4), write('.'),
        findall(a, solve_fc(four_queens), B), length(B, 2), write('.'),
        findall(a, solve_fc(n_queens(1)), N1), length(N1, 1), write('.'),
        findall(a, solve_fc(n_queens(2)), N2), length(N2, 0), write('.'),
        findall(a, solve_fc(n_queens(3)), N3), length(N3, 0), write('.'),
        findall(a, solve_fc(n_queens(4)), N4), length(N4, 2), write('.'),
        findall(a, solve_fc(n_queens(5)), N5), length(N5, 10), write('.'),
        findall(a, solve_fc(n_queens(6)), N6), length(N6, 4), write('.'),
        findall(a, solve_fc(n_queens(7)), N7), length(N7, 40), write('.'),
        findall(a, solve_fc(n_queens(8)), N8), length(N8, 92), write('.'),
        findall(a, solve_fc(n_queens(9)), N9), length(N9, 352), write('.'),
        findall(a, solve_fc(n_queens(10)), N10), length(N10, 724), write('.'),
        !, write('Exercițiul 4.2 rezolvate corect'), nl.


%% 5. BONUS [2 puncte]
%% Pornind de la predicatul forward_checking/3 de mai sus, scrieți predicatul
%% forward_checking_vo/3 astfel încât la fiecare pas al căutării să fie
%% selectată pentru legare variabila implicată în cele mai multe constrângeri.
%% Trebuie considerate doar constrângerile între variabile libere.

%% Implementați predicatul auxiliar get_most_constrained/3
%% get_most_constrained(Vars, Constraints, Var)

get_most_constrained(_Vars, _Constraints, _MCVar):-fail.

check_bonus:-
        get_most_constrained([X1,X2,X3],
                             [c([X1,X2],X1 < X2), c([X1,X3], X1 > X3)],
                             Var1),
        Var1 == X1, var(X1), write('.'),
        get_most_constrained([X1,X2,X3,X4],
                             [c([X1,X2],X1 < X2), c([X4,X1], X4 =:= X1),
                              c([X4,X2],X4 =< X2), c([X4,X3], X4 + X3 > 2)],
                             Var2),
        Var2 == X4, var(X4), write('.'),
        get_most_constrained([X1,X2,X3,X4],
                             [c([X1,2], X1 > 2), c([5,X1], X1 < 5),
                              c([X1,X2],X1 < X2), c([X4,X1], X4 =:= X1),
                              c([X4,X2],X4 =< X2), c([X4,X3], X4 + X3 > 2)],
                             Var3),
        Var3 == X4, var(X4), write('.'),
        !, write('Bonus rezolvat corect'), nl.

check_all:-
        check_12,
        check_3,
        check_41,
        check_42,
        check_bonus.